/*
 *  main.c - main() of the server (flcdd)
 *           This file is part of the FreeLCD package.
 *
 *  $Id: main.c,v 1.7 2003/11/04 23:13:01 unicorn Exp $
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Copyright (c) 2002, 2003, Jeroen van den Berg <unicorn@hippie.nu>
 */

#define GNU_PREFIX ""   /* "GNU " */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if defined (MSDOS) || defined (WIN32) || defined (_WIN64)
# include <io.h>  /* For close() */
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#else
# if HAVE_SYS_UNISTD_H
#  include <sys/unistd.h>
# endif
#endif

#if HAVE_LOCALE_H
# include <locale.h>
#endif

#if HAVE_SYSLOG_H
# include <syslog.h>
#else
# define syslog(FOO,BAR,BAZ)
#endif

#if HAVE_TYPES_H
# include <types.h>
#else 
# if HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
#endif

#if HAVE_STRING_H
# include <string.h>
#else
# if HAVE_STRINGS_H
#  include <strings.h>
# endif
#endif

#if USE_GUILE
# include <libguile.h>
#endif

#if STDC_HEADERS
# include <stdlib.h>
#else
extern void exit ();
extern char *malloc ();
#endif

#include <pwd.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <sys/stat.h>

#if HAVE_GETOPT_LONG
# include <getopt.h>
#endif

#if HAVE_SOCKET_H
# include <socket.h>
#else
# if HAVE_SYS_SOCKET_H
#  include <sys/socket.h>
# endif
#endif


#include "renderers/charlayout.h"
#include "renderers/charrenderer.h"
#include "renderers/charcanvas.h"

#include "common/socket.h"
#include "common/xmlt.h"

#include "connections.h"
#include "drivers.h"
#include "xmlconfig.h"
#include "scheduler.h"
#include "guile.h"

/* gettext helper macro */
#if ENABLE_NLS
# include <libintl.h>
# define _(foo) gettext (foo)
#else
# define bindtextdomain(X,Y)
# define textdomain(Domain)
# define _(foo) foo
#endif /* enable_nls */

/* Options definition */
#if HAVE_GETOPT_LONG
static struct option longopts[] = {
  {"port", required_argument, 0, 'p'},
# if HAVE_FORK
   {"foreground", no_argument, 0, 'f'},
# endif
  {"help", no_argument, 0, 'h'},
  {"version", no_argument, 0, 'V'},
  {0, 0, 0, 0}
};
#endif /* have getopt_long */

#if USE_GUILE
  void inner_main (void *, int, char **);
#endif

int keep_going;



/*----------------------------------------------------- sighup_handler --*/
static void
sighup_handler ()
{
  syslog (LOG_INFO, _("received SIGHUP"));
}

/*---------------------------------------------------- sigterm_handler --*/
static void
sigterm_handler ()
{
  syslog (LOG_INFO, _("received SIGTERM, exiting"));
  exit (EXIT_SUCCESS);
}

/*------------------------------------------------------- exit_program --*/
static void
exit_program ()
{
  syslog (LOG_INFO, _("exiting"));
  closelog ();
  exit (EXIT_FAILURE);
}

/*------------------------------------------------------------ sys_err --*/
static void
sys_err (const char *errstr)
{
  syslog (LOG_ERR, errstr);
  exit_program ();
}

/*---------------------------------------------------------- show_help --*/
static void
show_help (const char *app_name)
{
  printf (_("Usage: %s [options] ...\nOptions:\n"), app_name);
#ifdef HAVE_GETOPT_LONG
  printf (_(" \
--port, -p <portnr> Let the server listen on another port number\n"));
# if HAVE_FORK
  printf (_(" --foreground, -f    Run the server in the foreground\n"));
# endif
  printf (_(" --help, -h          Program usage\n"));
  printf (_(" --version, -V       Show program version and exit\n"));
#else /* ! HAVE_FORK */
  printf (_(" -p <portnr>  Let the server listen on another port number\n"));
# if HAVE_FORK
  printf (_(" -f           Run the server in the foreground\n"));
# endif
  printf (_(" -h           Program usage\n"));
  printf (_(" -V           Show program version and exit\n"));
#endif /* HAVE_FORK */
  printf (_("\nReport bugs to <%s>\n"), PACKAGE_BUGREPORT);
}

/*------------------------------------------------------- show_version --*/
static void
show_version (const char *app_name)
{
  printf ("%s - %s%s\n", app_name, GNU_PREFIX, PACKAGE_STRING);
  printf (_("Copyright (C) %s %s\n"), "2002, 2003", "Jeroen van den Berg");
  printf (_("\
This program is free software; you may redistribute it under the terms of\n\
the GNU General Public License.  This program comes with absolutely NO\n\
WARRANTY. For more information about these matters, see the file named\n\
COPYING.\n"));
}

/*------------------------------------------------ drop_root_privilege --*/
static void
drop_root_privilege (const char *user)
{
  struct passwd *pw_ent;

  /*  If we have root privs, try to set our uid to something safer.
   *  First try a specified username, then try "nobody" (which is a
   *  reasonable fallback on most systems).
   */
  if (getuid () == 0 || geteuid () == 0)
    {
      pw_ent = getpwnam (user);
      if (pw_ent == NULL)
        {
          syslog (LOG_WARNING,
                  _("user %s is not a valid user, fallback to 'nobody'\n"),
                  user);
          pw_ent = getpwnam ("nobody");
          if (pw_ent == NULL)
            {
              sys_err (_("cannot drop root privileges, aborting"));
              exit (EXIT_FAILURE);
            }
        }

      if (setuid (pw_ent->pw_uid) < 0)
        sys_err ("setuid");
    }
}


/*--------------------------------------------------------------- main --*/
int
main (int argc, char **argv)
#if USE_GUILE
{
  scm_boot_guile (argc, argv, inner_main, 0);
  return EXIT_SUCCESS;
}
        
void 
inner_main (void *closure, int argc, char **argv)
#endif
{
  char         c;
  int          run_in_foreground = 0; /* (--foreground) */
  unsigned int port = 8800; /* (--port) */
  const char   *app_name;
  pid_t        child; 
  long         tmp;
  socket_t     sock;
  fd_set       fdset;
  xml_node     *xconf;
  xml_node     *output_config;
  void         *driver;
  
#if HAVE_SETLOCALE
  /* Set up gettext internationalisation */
  if (setlocale (LC_ALL, "") == NULL)
    {
      perror (_("warning: setting locale to LC_ALL failed"));
    }
#endif

#if ENABLE_NLS
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
#endif

  /* Find out name of executable */
  app_name = *argv + strlen (*argv) - 1;
  while (app_name > *argv && *(app_name - 1) != '/')
    --app_name;

  /* Parse command line options */
#ifdef HAVE_GETOPT_LONG
  while ((c = getopt_long (argc, argv, "p:bhV", longopts, 0)) > 0)
#else
  while ((c = getopt (argc, argv, "p:bhV")) > 0)
#endif
    {
      switch (c)
        {
        case 'p':                 /* --port, -p */
          tmp = strtol (optarg, 0, 10);
          if (errno == ERANGE || errno == EINVAL || tmp < 1 || tmp > 65536)
            {
              fprintf (stderr,
                _("%s: Port number must be between 1 and 65536\n"), app_name);
              exit (EXIT_FAILURE);
            }
          port = (unsigned int)tmp;
          break;

        case 'f':                /* --foreground, -f */
          run_in_foreground = 1;
          break;
              
        case 'h':                /* --help, -h */
          show_help (app_name);
          exit (EXIT_SUCCESS);

        case 'V':                /* --version, -V */
          show_version (app_name);
          exit (EXIT_SUCCESS);

        case '?':               /* Invalid option */
          show_help (app_name);
          exit (EXIT_FAILURE);

        default:
          /* In an ideal world, this shouldn't happen. */
          exit (EXIT_FAILURE);
        }
    }

  if (optind < argc)
    {
      perror (_("too many arguments"));
      exit (EXIT_FAILURE);
    }

  /* Fork into the background */
#ifdef HAVE_FORK
  if (!run_in_foreground)
    {
      child = fork();
      if (child < 0)
        sys_err ("fork");
            
      if (child != 0)
        exit (EXIT_SUCCESS);

      /* Create a new session */
      if (setsid() < 0 )
        sys_err ("setsid");
    }
#endif

  /* Configure system logging */
  openlog (app_name, LOG_CONS | LOG_PID | LOG_PERROR, LOG_DAEMON);

  /* Take some precautions */
  /* FIXME: uncomment this part when configuration files are handled properly.
  if (chdir( "/" ) < 0)
    sys_err ("chdir");
  */
  umask (0);
  
  if (!run_in_foreground)
    {
      close (STDOUT_FILENO);
      close (STDERR_FILENO);
    }
  close (STDIN_FILENO);

  /* Open server socket */
  sock = socket_create_server (port, &fdset);
  if (sock < 0)
    {
      switch (sock)
        {
        case -1:
          sys_err ("socket()");
          break;
        case -2:
          sys_err ("setsockopt()");
          break;
        case -3:
          sys_err ("inet_aton()");
          break;
        case -4:
          sys_err ("bind()");
          break;
        default:
          sys_err (_("cannot open server socket"));
          break;
        }
      exit (EXIT_FAILURE);
    }

  /* We do not need "dangerous" privileges anymore, drop them before
   * messing with potential dangers like plug-ins. 
   */
  drop_root_privilege (app_name);

  /* Initialise screen scheduling. */
  sch_init ();
  
  /* Parse the configuration file. For now, we look in the current directory.
   * Somewhere in the future this should become /etc or ~ . 
   */
  xconf = xc_parse_config_file ("flcdd.xml");
  if (xconf == NULL)
    {
      sys_err (_("Failed to parse inifile"));
      exit (EXIT_FAILURE);
    }

  /* Load driver modules */
  if (drivers_init () != 0)
    {
      sys_err (_("Failed to initialise drivers"));
      exit (EXIT_FAILURE);
    }
  
  output_config = 0;
  while ((output_config = xmlt_find (xconf, output_config, OUTPUT)) != 0)
    {
      const char *name = xmlt_get_attrib (output_config, NAME);

      if (name == NULL)
        {
          puts ("No output name given.");
          continue;
        }
      
      driver = drivers_create_output (output_config);

      if (driver == NULL)
        {
          puts ("drivers_create_output failed");
        }
      else
        {
          sch_register_driver (driver, name);
        }
    }

  xmlt_free_document (xconf);

  /* If necessary, run the Guile initialisation files. */
  if (!guile_init ())
    {
      sys_err (_("Failed to initialise libguile"));
      exit (EXIT_FAILURE);
    }
  
  if (signal (SIGHUP, sighup_handler) == SIG_ERR)
    {
      sys_err (_("Cannot catch SIGHUP"));
      exit (EXIT_FAILURE);
    }
  
  if (signal (SIGTERM, sigterm_handler) == SIG_ERR)
    {
      sys_err (_("Cannot catch SIGHUP"));
      exit (EXIT_FAILURE);
    }

  /* Enter event loop */
  syslog (LOG_INFO, _("running"));
  keep_going = 1;
  while (keep_going != 0)
    {
      /*  Poll the sockets from our file descriptor set, and call
       *  one of the conn_* functions whenever something interesting
       *  happens.
       */
      socket_poll (sock, &fdset, conn_open, conn_data, conn_close);
    }

  closelog ();

#if !USE_GUILE  
  return EXIT_SUCCESS;
#endif
}
